# File format:
# 
# [score_type lowscore|highscore]
# [game_name <name>]
# 
# game <time>
# <player> <score>
# <player> <score>
# <empty line>
#
# <etc>
# Hash comments are OK
#
# Example:
# 
# game 2010-08-02 20:12
# Peter      100
# Jonatan    210
# Axel       134
#
#
# After parsing we have a list of dicts that look like this:
# { 'time': String, 'scores': dict }
# The scores dict is a mapping between player names and their scores
# in that game.

import re
import sys
import math


class UnexpectedStatementError(Exception): pass


class GamesParser(object):
    def __init__(self, file_obj, name=None):
        self.file = file_obj
        self.score_type = 'highscore'
        self.game_name = name
        self.errors = []

    def parse_type(self, line):
        type_exp = '^score_type\s+(lowscore|highscore)$'
        m = re.match(type_exp, line)
        if m is not None:
                self.score_type = m.group(1)
                return True
        else:
            return False

    def parse_game_name(self, line):
        name_exp = '^game_name\s+(.+)$'
        m = re.match(name_exp, line)
        if (m is not None):
                self.game_name = m.group(1)
                return True
        else:
            return False

    def parse_game(self, line):
        game_exp = '^game\s+([0-9]{4}-[0-9]{2}-[0-9]{2}\s+[0-9]{2}[:.][0-9]{2})\s*$'
        m = re.match(game_exp, line)
        if m is not None:
            time_string = m.group(1)
            return { 
                'time': time_string,
                'scores': {}
                }
        else:
            return None

    def parse_score(self, line, game_dict):
        score_exp = '^([a-zA-Z\+]+)\s+(-?[0-9]+)\s*$'
        m = re.match(score_exp, line)
        if m is not None:
            if game_dict is None:
                raise UnexpectedStatementError('Spurious score (no current game) on line %i' % self.line_no)
            else:
                name = m.group(1)
                score = int(m.group(2))
                game_dict['scores'][name] = score
                return True
        else:
            return False

    def parse_comment(self, line):
        return re.match('^#.*$', line) is not None

    def parse_empty_line(self, line):
        return re.match('^\s*$', line) is not None

    def parse_file(self):
        self.games = []
        current_game = None
        self.line_no = 0

        for line in self.file:
            self.line_no += 1
            try:
                if self.parse_comment(line) or self.parse_empty_line(line):
                    continue
                
                if self.parse_type(line):
                    continue

                if self.parse_game_name(line):
                    continue

                if self.parse_score(line, current_game):
                    continue

                new_game = self.parse_game(line)
                if new_game:
                    if current_game is not None:
                        self.games.append(current_game)
                    current_game = new_game
                    continue
    
                raise UnexpectedStatementError('Syntax error on line %i: %s' % (self.line_no, line))
            except UnexpectedStatementError as e:
                self.errors.append(e.message)
        
        if current_game is not None:
            self.games.append(current_game)

        return self.games



import math
import time

from player import Player

#def to_lowscore(score, scores):
#    max_score_in_game = max(scores)
#    return max_score_in_game - score

def to_lowscore(score, scores):
    max_score_in_game = max(scores)
    min_score_in_game = min(scores)
    return max_score_in_game - score + min_score_in_game

class Rater(object):
    def __init__(self, games):
        self.games = games
        self.players = {}


    def calculate_new_rating(self, score_type, player, game):
        return self.calculate_new_rating_keep_average(score_type, player, game)

    def calculate_new_rating_betapet(self, score_type, player, game):
        """
        Calculates new rating for a player based on old rating and result in the 
        given game. See http://betapet.com/rating for an explanation.
        """
        square_scores = True
        use_diminishing = False

        old_rating = self.players[player].get_rating()
        # Using max to make 0 the lowest possible score
        score = max(game['scores'][player], 0)
        scores = game['scores'].values()

        if score_type == 'lowscore':
            score = to_lowscore(score, scores)
            scores = [to_lowscore(s, scores) for s in scores]

        if square_scores:
            # square scores
            score = score ** 2
            scores_sum = math.fsum(max(s, 0) ** 2 for s in scores)
        else:
            scores_sum = math.fsum(max(s, 0) for s in scores)
            
        # TODO: Need to handle scores_sum == 0. Check betapet... 
        old_ratings_sum = math.fsum(self.players[p].get_rating() for p in game['scores'].keys())

        games = self.players[player].get_game_count()
        if use_diminishing:
            diminishing_part = 0.2 * math.exp(-0.1 * games)
        else:
            diminishing_part = 0

        change = ((score / scores_sum - 
                   old_rating / old_ratings_sum) * 
                  old_rating * 
                  (0.1 + diminishing_part))
        new_rating = old_rating + change
        
        return new_rating


    def calculate_new_rating_keep_average(self, score_type, player, game):
        """
        Calculates new rating for a player based on old rating and result in the 
        given game. The average rating will be the same before and after this 
        calculation has been called for all players.
        """
        
        old_rating = self.players[player].get_rating()
        # Using max to make 0 the lowest possible score
        score = max(game['scores'][player], 0)
        scores = game['scores'].values()

        if score_type == 'lowscore':
            score = to_lowscore(score, scores)
            scores = [to_lowscore(s, scores) for s in scores]

        scores_sum = math.fsum(max(s, 0) for s in scores)
        old_ratings_sum = math.fsum(self.players[p].get_rating() for p in game['scores'].keys())
            
        K = 0.05
        rating_in = K * old_rating
        rating_out = (score / scores_sum) * K * old_ratings_sum
        change = rating_out - rating_in
        new_rating = old_rating + change

        return new_rating


    def cmp_rating(self, p1, p2):
        if self.players[p1].get_rating() < self.players[p2].get_rating():
            return -1
        elif self.players[p1].get_rating() > self.players[p2].get_rating():
            return 1
        else:
            return 0
        
    def rate_games(self, score_type): 
        game_num = 0
        for game in self.games:
            game_num += 1
            # Add any new players to the players dict
            for player in game['scores']:
                if player not in self.players:
                    self.players[player] = Player(player)
                   
            # Calculate ratings
            new_ratings = {}
            for player in game['scores']:
                new_ratings[player] = self.calculate_new_rating(score_type, player, game)
            
           
            # Update ratings
            for player in new_ratings:
                self.players[player].set_rating(new_ratings[player], game_num)

            #rats = [self.players[p].get_rating() for p in self.players]
            #print math.fsum(rats) / len(rats)


        sorted_players = self.players.keys()
        sorted_players.sort(self.cmp_rating, reverse=True) 

        ratings = []
        for player in sorted_players:
            ratings.append((player, 
                            self.players[player].get_game_count(), 
                            round(self.players[player].get_rating()),
                            round(self.players[player].get_rating_change(game_num))))
            
        return ratings



#!/usr/bin/python

import sys
import json
from game_parser import GamesParser
from rater import Rater
from optparse import OptionParser

def parse_file(filename, mode="text", player_name=None):
    out = []
    
    try:
        with file(filename) as f:
            parser = GamesParser(f, filename)
            games = parser.parse_file()
            for error in parser.errors:
                print error

        rater = Rater(games)
        ratings = rater.rate_games(parser.score_type)
        

        if mode == "gnuplot":
            commands = 'set key left\n'
            plot_command = 'plot "%s" with lines lw 2 title "%s"\n'
            
            for name in rater.players:
                player = rater.players[name]
                if player.get_game_count() > 0:
                    player_file_name = "plot_rating_%s.data" % player.name
                    commands += plot_command % (player_file_name, player.name)
                    if plot_command.startswith('plot'):
                        plot_command = 're' + plot_command
                    player_file = open(player_file_name, "w")  
                    for game_num in range(player.get_first_game() - 1, len(games) + 1):
                        rating = player.get_rating(game_num)
                        print >> player_file, "%d %d" % (game_num, rating)
    
            commands += 'pause mouse\n'
            command_file_name = "plot_commands.gnuplot"
            command_file = open(command_file_name, "w")
            print >> command_file, commands
            out.append("Run 'gnuplot %s' to plot all ratings." % command_file_name)
        elif mode == "json":
            rating_dict = {}
            for rating in ratings:
                r = {"name":rating[0], "matches":rating[1], "rating":rating[2], "delta":rating[3]}
                rating_dict[rating[0]] = r
            out.append(json.dumps(rating_dict))
        elif player_name is not None:
            player = rater.players[player_name]
            if player:
                for game_num in sorted(player.rating_history.keys()):
                    rating = player.rating_history[game_num]
                    out.append("%d %d" % (game_num, rating))
        else:
            out.append("Game: %s" % parser.game_name)
            out.append("Total number of games: %s" % len(games))
            hformat = "%-15s%-10s%-10s%-10s"
            rformat = "%-15s%7d%9d%9d"
            out.append(hformat % ("Name", "Games", "Rating", "Delta"))
            for rating in ratings:
                out.append(rformat % rating)
    except IOError as ioe:
        out.append(str(ioe))
    
    return "\n".join(out)

def main():
    opt_parser = OptionParser(usage = "usage: %prog [options] <game file>")
    opt_parser.add_option("-m", "--mode", dest="mode", default="text",
                      help="output mode (text, html, gnuplot, json)")
    opt_parser.add_option("-p", "--player", dest="player",
                      help="player name")
    
    (options, args) = opt_parser.parse_args()
    
    if len(args) < 1:
        print >> sys.stderr, "Error: No game file specified. Try", sys.argv[0], "-h."
        exit(1)
    
    print parse_file(args[0], options.mode, options.player)


if __name__ == '__main__':
    main()



START_RATING = 1000

class Player(object):
    def __init__(self, name):
        self.rating_history = {}
#        self.set_rating(START_RATING, 0)
        self.name = name

    def get_rating_change(self, game):
        if self.get_game_count() > 0 and game in self.rating_history:
            return self.get_rating(game) - self.get_rating(game - 1)
        else:
            return 0

    def has_played(self, game):
        return game in self.rating_history

    def get_rating(self, game = None):
        if self.get_game_count() is 0:
            game = 0
        elif game is None:
            game = self.last_game
        elif game < 0:
            game = 0

        if not self.has_played(game):
            if game is not 0:
                return self.get_rating(game - 1)
            else:
                return START_RATING
        else:
            return self.rating_history[game]
    
    def set_rating(self, rating, game_num):
        self.rating_history[game_num] = rating
        self.last_game = game_num

    def get_game_count(self):
        return len(self.rating_history)

    def get_first_game(self):
        if self.get_game_count() > 0:
            return sorted(self.rating_history.keys())[0]
        else:
            return None



# encoding: utf-8

from setuptools import setup, find_packages
import sys, os

version = '0.1.4'
 
setup(
    name='rayter',
    version=version,
    description="Game rating command line tool",
    long_description="""Game rating command line tool""",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Games/Entertainment",
    ],
    keywords='',
    author='Peter Jaric, Jonatan Heyman',
    author_email='peter@jaric.org, jonatan@heyman.info',
    url='https://github.com/peterjaric/rayter',
    license='MIT',
    platform='Any',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    include_package_data=True,
    zip_safe=False,
    install_requires=[],
    entry_points={
        'console_scripts': [
            'rayter = rayter.main:main',
        ]
    },
    #test_suite='rayter.test.runtests',
)
